
// 引入node crypto加解密模块，用于完整性校验
const crypto = require("crypto");
const Buffer = require("buffer");
const path = require("path");

/** 插件默认配置对象
 * @pluginName 自定义插件名称
 * @injectName 输出的映射资源注入到window的变量名称，或者映射文件名字
 * @chunksMapFileFun 自定义输出的映射资源文件的名称
 *   contentHash 用于完整性校验，内容变动都会改变，更好利用缓存
 * @chunkNameFilter chunk输出过滤器：哪些chunk需要生成映射
 *    chunkName 参数为chunk.name
 *    @return boolean true时输出
 * @chunkKeyFun 自定义输出的映射文件中的json对象里的key
 *    chunkName
 *    @return chunkName
 * @injectToHtmlFilter 要注入到的HTML文件的，HTML文件资源过滤器
 *    assetName 打包生成的资源字符串，默认index.html文件
 *    @return boolean true时注入
 * @assetsFilter assets资源输出过滤器
 *    assetUrl
 *    @return boolean true时输出
 * @assetToChunkRegs 把指定资源，映射到指定chunk中，可配置多个筛选项
 *    chunkNameFilter 指定chunkname过滤器
 *    assetNameFilter 指定asset资源过滤器
 *
 * 把包含link-layout的资源放入，包含link的chunkname对应的映射中
 * [{
    chunkNameFilter:(chunkName)=>/link/.test(chunkName),
    assetNameFilter:(assetUrl)=>/link-layout/.test(assetUrl)}]
 */

const defaultOptions = {
  pluginName: "webpack-prefetch-plugin",
  injectName: "_chunks_map",
  chunksMapFileFun: contentHash =>`${this.injectName}.${contentHash}.js`,
  // chunkNameFilter: chunkName => !!chunkName,
  chunkNameFilter: (chunkName)=> chunkName.includes('_'),
  chunkKeyFun: chunkName => chunkName,
  injectToHtmlFilter: assetName => /^index\.html?$/i.test(assetName),
  assetsFilter: assetUrl => {
    // 默认过滤掉chunk对应的映射中的map文件
    return !["map"].includes(path.extname(assetUrl));
  },
  // [{ chunkNameFilter:(chunkName)=>true,assetNameFilter:(assetUrl)=>true}]
  assetToChunkRegs: [],
};

/**
 * 插件定义，使用类时，必须要有apply方法
 */
class ChunksMapPlugin {
  constructor(options = {}) {
    // 覆盖默认options
    this.options = Object.assign({}, defaultOptions, options);
  }
  // webpack调用apply方法，传入webpack compiler编译器实例
  apply(compiler) {
    const _this = this;

    /** 即将输入生成的文件时，触发emit钩子，tapAsync类型
     * @pluginName 插件名称
     * @callpack 回调函数
     *    complication 保存着本次编译的资源、各种方法
     *    next 处理完自身逻辑，需要调用next()，webpack继续执行
     */
    compiler.hooks.emit.tapAsync(
      _this.options.pluginName,
      function (compilation, next) {
        // 要输出的资源资源对象，类似：
        // { _home_page:['_home_page.232.js','_home_page.232.css'] }
        const chunksMap = {};

        // 根据assetNameFilter函数筛选出特定资源，放入该规则的assets属性
        const assetToChunkRegs = _this.options.assetToChunkRegs.map(
          regObject => {
            // 根据assets Key即资源名称，过滤出需要映射的chunk依赖的资源
            const assets = Object.keys(compilation.assets).filter(item =>
              regObject.assetNameFilter(item)
            );
            return Object.assign({}, regObject, {  assets });
          }
        );

        /** compilation
         * @chunks 存放所有打包生成的chunk代码块
         *    files 由chunk生成的文件，一般拆分成js,css,map文件,修改该数组对打包结果没有影响
         *    modulesIterable 依赖模块迭代器
         * @assets 存放即将要输入的资源对象
         */
        compilation.chunks.forEach(function (chunk) {
          // 当前chunk对应的资源数组,包含由chunk生成assets,以及chunk的依赖模块生成的assets
          let files = [];

          // 筛选出路由对应的chunk，生成的文件
          if (_this.options.chunkNameFilter(chunk.name)) {
            // 如果当前chunk需要生成映射

            // 保存由chunk生成assets文件
            files = chunk.files || [];

            for (const regObject of assetToChunkRegs) {
              // 找出满足chunkNameFilter规则的资源也加入映射
              if (regObject.chunkNameFilter(chunk.name)) {
                files = files.concat(regObject.assets);
              }
            }

            /**
             * module.buildInfo 该模块打包信息
             * buildInfo.assets 打包后生成的assets对象
             */
            for (const module of chunk.modulesIterable) {
              if (module.buildInfo && module.buildInfo.assets) {
                // 把该chunk的依赖模块资源也加入映射中
                files = files.concat(Object.keys(module.buildInfo.assets));
              }
            }

            // chunk名称作为映射对象的每个key
            const chunkKey = _this.options.chunkKeyFun(chunk.name);

            // 保存chunk和chunk筛选后的对应资源的映射
            chunksMap[chunkKey] = files.filter(function (url) {
              return _this.options.assetsFilter(url);
            });
          }
        });

        // 把生成的资源映射对象转成JSON字符串
        const chunksMapString = JSON.stringify(chunksMap);

        // 根据资源映射字符串生成内容摘要
        const contentHash = _this.generateHashByContent(chunksMapString);

        // 拼接JavaScript脚本字符串，把资源映射，放入全局变量中
        const configJs =
          "\n(typeof window=='undefined'?global:window)." +
          _this.options.injectName +
          "=" +
          chunksMapString +
          ";\n";

        // 获取资源映射的内容要输出的文件名字
        const outputFileName = _this.options.chunksMapFileFun(contentHash);

        // 手动添加要输出映射文件，在assets中用对象表示
        compilation.assets[outputFileName] = {
          name:outputFileName,
          source: () => configJs,
          size: () => {
            return Buffer.byteLength(configJs, "utf8");
          },
        };

        _this.addToHtmlFun(compilation, _this.options.injectToHtmlFilter, outputFileName);
        // 这是一个异步事件，要记得调用 callback 通知 Webpack 本次事件监听处理结束。
        // 如果忘记了调用 callback，Webpack 将一直卡在这里而不会往后执行。
        next();
      }
    );
  }

  /**
   * 根据内容生成 contenthash，如果内容没变生成的contenthash是一样的
   * 修复内容改变，继续使用旧缓存的问题，此时需要获取最新的映射资源
   */
  generateHashByContent(content) {
    if (typeof content !== "string" && !Buffer.isBuffer(content)) {
      throw new TypeError("Expected a Buffer or string");
    }

    /** 使用md5（sha1，sha256默认，sha512）算法生成一个hash实例
     * hash.update('要加密的字符串')，返回hash实例
     * hash.digest('hex) 生成内容摘要。默认是2进制，转成16进制
     * 调用String.slice()截取前十位
     */
    return crypto.createHash("md5").update(content).digest("hex").slice(0, 10);
  }

  // 直接注入脚本代码到HTML文件
  addToHtmlFun(compilation, injectToHtmlFilter, outputFileName) {

    // 从将要输出的资源中，找出index.html文件
    const indexHtmls = Object.keys(compilation.assets).filter(assetName => {
      return injectToHtmlFilter(assetName);
    });

    indexHtmls.forEach(htmlName => {
      const htmlSource = compilation.assets[htmlName];
      // 获取index.html文件的内容
      const htmlContent = htmlSource.source();

      if (htmlContent) {
        // 把生成的映射文件，添加进index.html文件的 body元素最后
        htmlContent = htmlContent.replace(
          /\<\/body\>/i,
          m => `<script src=${outputFileName} async></script>${m}`
        );
      }

      // 把修改后的html，手动添加到要输入文件中
      delete compilation.assets[htmlName];
      compilation.assets[htmlName] = {
        name:htmlName,
        source: () => htmlContent,
        size: () => {
          return Buffer.byteLength(htmlContent, "utf8");
        },
      };
    });
  }
}
module.exports = ChunksMapPlugin;